/*  Screem:  editMenu.c,
 *  This file handles find/replace
 *
 *  Copyright (C) 2000  David A Knight
 *
 *  TODO:
 *  
 *  ability to find backwards as well as forward?
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */

#include <config.h>
#include <sys/types.h>

#include <regex.h>

#include <libgnome/gnome-i18n.h>
#include <libgnomeui/gnome-entry.h>

#include <libgnomevfs/gnome-vfs-utils.h>

#include <gtk/gtkdialog.h>
#include <gtk/gtktreeview.h>
#include <gtk/gtktogglebutton.h>
#include <gtk/gtkcellrenderer.h>
#include <gtk/gtkcellrenderertext.h>
#include <gtk/gtkliststore.h>

#include <gconf/gconf-client.h>

#include <glade/glade.h>

#include <string.h>

#include "editMenu.h"
#include "pageUI.h"

#include "screem-window.h"
#include "screem-window-private.h" /* YUK */
#include "screem-site.h"
#include "screem-page.h"
#include "screem-editor.h"

#include "screem-file-browser.h"

static void search_all_pages( GtkWidget *widget );
static gint search( const gchar *text, const gchar *find, gboolean use_regex,
		    gint start, gint *length );

void do_find( GtkWidget *widget );
void do_replace( GtkWidget *widget );
void do_replace_all( GtkWidget *widget );

void find_change_file( GtkTreeView *view, GtkTreePath *path,
		       GtkTreeViewColumn *col, gpointer data );

/**
 * find_text:
 *
 * @string:  the string to search
 * @find:    the string to search for
 * @replace: the string to replace find with, NULL if not replacing
 * @length:  a pointer to the place to store the length of the match,
 *           NULL if not wanted.
 *
 * Find the string find in string, and replace with replace if given,
 * storing the length of the match in length, if given
 *
 * return: a pointer to the found position, or the new text if a replace was
 *         performed
 */
gchar *find_text( const gchar *string, const gchar *find, 
		  const gchar *replace, gint *length )
{
	gchar *chunk1;
	gchar *chunk2;
	gchar *newChunk;
	gchar *text = NULL;
	gint found_at;
	gint find_length;

       	found_at = search( string, find, TRUE, 0, &find_length );
    
	if( found_at == -1 ) {
		return NULL;
	}

	/* replace the text if needed */
	if( replace ) {
		chunk1 = g_strndup( string, found_at );
		chunk2 = g_strdup( string + found_at + find_length );
		newChunk = g_strconcat( chunk1, replace, chunk2, NULL );
		g_free( chunk1 );
		g_free( chunk2 );
		text = newChunk;
	} else {
		text = (gchar*)string + found_at;
	}
      
	if( length ) {
		*length = find_length;
	}

	return text;
}


/**
 * screem_tags_to_upper:
 *
 * proxy to call to screem_tags_case_change with the correct flag
 *
 * returns: none
 */
void screem_tags_to_upper()
{
	screem_tags_case_change( TRUE );
}

/**
 * screem_tags_to_lower:
 *
 * proxy to call to screem_tags_case_change with the correct flag
 *
 * returns: none
 */
void screem_tags_to_lower()
{
	screem_tags_case_change( FALSE );
}

/**
 * screem_tags_case_change:
 *
 * @upper: TRUE if the change is to uppercase, FALSE if it is to lowercase.
 *
 * Changes the case of all tags and attributes in the current document
 *
 * returns: none
 */
void screem_tags_case_change( gboolean upper )
{
	/*	ScreemPage *page;
	gchar *text;

	gint start = 0;
	gint len = -1;


	if( screem_editor_has_selection( &start, &len ) )
		len = len - start;
    
	text = screem_editor_get_text( start, len );
	screem_markup_change_case( text, upper );
	screem_editor_delete_forward( start, len );
	screem_editor_insert( start, text );
	g_free( text );*/
}

/**
 * find:
 *
 * proxy to call find replace with the correct flag
 *
 * returns: none
 */
void screem_find( ScreemWindow *window )
{
	screem_find_replace( window, FALSE );
}

/**
 * replace:
 *
 * proxy to call find replace with the correct flag
 *
 * returns: none
 */
void screem_replace( ScreemWindow *window )
{
	screem_find_replace( window, TRUE );
}

/**
 * find_replace:
 *
 * @replace:   boolean flag to indicate if the replace features should
 *             be made sensitive.
 *
 * displays the find/replace dialog
 *
 * returns: none
 */
void screem_find_replace( ScreemWindow *window, gboolean replace )
{
	GladeXML *xml;
	GtkWidget *widget;
	gchar *glade_path;
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *col;
	GtkListStore *store;
	GConfClient *client;

	client = gconf_client_get_default();

	glade_path = gconf_client_get_string( client,
					      "/apps/screem/general/glade_path",
					      NULL );
	g_object_unref( client );
	xml = glade_xml_new( glade_path, "find_dialog", NULL );
	g_free( glade_path );

	widget = glade_xml_get_widget( xml, "find_dialog" );
	g_object_set_data( G_OBJECT( widget ), "window", window );
	gtk_window_set_transient_for( GTK_WINDOW( widget ), GTK_WINDOW( window ) );

	gtk_dialog_set_default_response( GTK_DIALOG( widget ), 2 );
	
	widget = glade_xml_get_widget( xml, "find_entry" );
	widget = gnome_entry_gtk_entry( GNOME_ENTRY( widget ) );
	gtk_entry_set_activates_default( GTK_ENTRY( widget ), TRUE );
	
	/* setup the match list correctly */
	widget = glade_xml_get_widget( xml, "match_list" );

	col = gtk_tree_view_column_new();
	gtk_tree_view_column_set_title( col, "File" );
		
	renderer = gtk_cell_renderer_pixbuf_new();
	gtk_tree_view_column_pack_start( col, renderer, FALSE );
	gtk_tree_view_column_set_attributes( col, renderer, "pixbuf",
					     2, NULL );
	       
	renderer = gtk_cell_renderer_text_new();
	gtk_tree_view_column_pack_start( col, renderer, TRUE );
	gtk_tree_view_column_set_resizable( col, TRUE );
	gtk_tree_view_append_column( GTK_TREE_VIEW( widget ), col );
	gtk_tree_view_column_set_attributes( col, renderer, "text",
					     0, NULL );

	store = gtk_list_store_new( 3, 
				    G_TYPE_STRING, G_TYPE_STRING,
				    GDK_TYPE_PIXBUF,
				    NULL );
	gtk_tree_view_set_model( GTK_TREE_VIEW( widget ),
				 GTK_TREE_MODEL( store ) );
	
	g_object_set_data( G_OBJECT( widget ), "matches", NULL );

	g_signal_connect( G_OBJECT( widget ), "row_activated",
			  G_CALLBACK( find_change_file ), window );

	/* set last found position to -1 */
	widget = glade_xml_get_widget( xml, "replace_entry" );
	g_object_set_data( G_OBJECT( widget ), "found_at",
			   GINT_TO_POINTER( -1 ) );

	glade_xml_signal_autoconnect( xml );

	/* toggle the replace button if we are opened for replacing */
	widget = glade_xml_get_widget( xml, "replace_check" );
	gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON( widget ), replace );

	widget = glade_xml_get_widget( xml, "find_dialog" );
	gtk_dialog_run( GTK_DIALOG( widget ) );
}

void screem_find_replace_response( GtkWidget *widget, gint response, gpointer data )
{
	switch( response ) {
	case 0:
		do_replace( widget );
		break;
	case 1:
		do_replace_all( widget );
		break;
	case 2:
		do_find( widget );
		break;
	default:
		gtk_widget_destroy( widget );
		break;
	}
}

/**
 * search_all_pages:
 *
 * @widget:
 *
 * performs a single find on all pages in the site, and adds the
 * page to the found list if a match is found
 *
 * returns: none
 */
static void search_all_pages( GtkWidget *widget )
{
	GladeXML *xml;
	ScreemPage *page;
	GSList *list;
	GList *pages;
	GSList *matches;
	gchar *text;

	const gchar *find_string;
	const gchar *path;

	ScreemWindow *window;
	ScreemSite *site;
	GtkTreeModel *store;
	
	static ScreemFileBrowser *browser = NULL;
	
	if( ! browser ) {
		browser = screem_file_browser_new();
	}
	
	xml = glade_get_widget_tree( widget );
	
	widget = glade_xml_get_widget( xml, "find_dialog" );
	window = SCREEM_WINDOW( g_object_get_data( G_OBJECT( widget ),
						   "window" ) );
	site = screem_window_get_current( window );

	widget = glade_xml_get_widget( xml, "all_current" );
	if( gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( widget ) ) ) {
		pages = screem_site_get_pages( site );
	} else {
		pages = screem_window_get_documents( window );
	}
	
	widget = glade_xml_get_widget( xml, "find_entry" );
	widget = gnome_entry_gtk_entry( GNOME_ENTRY( widget ) );
	find_string = gtk_entry_get_text( GTK_ENTRY( widget ) );

	/* find pages with matches */
	for( matches = NULL; pages; pages = pages->next ) {
		page = SCREEM_PAGE( pages->data );

		path = screem_page_get_pathname( page );

		/* is path excluded? */
		if( ! screem_site_is_excluded( site, path ) ) {
			screem_page_load( page );
			text = screem_page_get_data( page );
			if( find_text( text, find_string, NULL, NULL ) ) {
				/* add to list */
				matches = g_slist_prepend( matches, page );
			}
			g_free( text );
		}
	}

	/* display matches */
	widget = glade_xml_get_widget( xml, "match_list" );
	
	store = gtk_tree_view_get_model( GTK_TREE_VIEW( widget ) );
	
	gtk_list_store_clear( GTK_LIST_STORE( store ) );
	
	for( list = matches; list; list = list->next ) {
		GtkTreeIter it;
		GdkPixbuf *pixbuf;
		gchar *temp;

		path = screem_page_get_pathname( SCREEM_PAGE( list->data ) );	
		temp = gnome_vfs_unescape_string_for_display( path );
		
		pixbuf = screem_file_browser_get_icon( browser, path, 16, 16,
							FALSE );

		gtk_list_store_append( GTK_LIST_STORE( store ), &it );
		
		gtk_list_store_set( GTK_LIST_STORE( store ), &it,
				    0, temp, 
				    1, find_string, 
				    2, pixbuf, -1 );
		g_free( temp );
	}
	
	g_object_set_data( G_OBJECT( widget ), "matches", matches );
}

/**
 * do_find:
 *
 * @widget: the find next button
 *
 * performs the next find
 *
 * returns: none
 */
void do_find( GtkWidget *widget )
{
	GladeXML *xml;
	const gchar *find_string;
	gboolean use_regex;

	gchar *text;
	gint found_at;
	gint length;
	gint pos;

	gboolean check_all;

	ScreemWindow *window;
	ScreemEditor *editor;
	ScreemPage *page;

	xml = glade_get_widget_tree( widget );

	widget = glade_xml_get_widget( xml, "find_dialog" );
	window = SCREEM_WINDOW( g_object_get_data( G_OBJECT( widget ),
						   "window" ) );
	editor = SCREEM_EDITOR( window->details->editor );
	page = screem_window_get_document( window );

	/* if there is no match list, and the user wants to check all
	   pages then check them */
	widget = glade_xml_get_widget( xml, "match_list" );
	if( ! g_object_get_data( G_OBJECT( widget ), "matches" ) ) {
		widget = glade_xml_get_widget( xml, "check_all" );
		check_all = gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(widget) );
		if( check_all )
			search_all_pages( widget );
	}

	widget = glade_xml_get_widget( xml, "find_entry" );
	widget = gnome_entry_gtk_entry( GNOME_ENTRY( widget ) );
	find_string = gtk_entry_get_text( GTK_ENTRY( widget ) );

	widget = glade_xml_get_widget( xml, "regex" );
	use_regex = gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON(widget) );

	widget = glade_xml_get_widget( xml, "pos_start" );
	pos = 0;
	if( ! gtk_toggle_button_get_active( GTK_TOGGLE_BUTTON( widget ) ) )
		pos = screem_editor_get_pos( editor );

	widget = glade_xml_get_widget( xml, "replace_entry" );
	found_at = (gint)g_object_get_data( G_OBJECT( widget ), "found_at" );
	if( found_at != -1 ) {
		pos = found_at + 1;
	}

	text = screem_page_get_data( page );

	found_at = search( text, find_string, use_regex, pos, &length );

	g_object_set_data( G_OBJECT( widget ), "found_at",
			   GINT_TO_POINTER( found_at ) );

	if( found_at != -1 ) {
		/* goto and highlight the string */
		screem_editor_set_pos( editor, found_at );
		screem_editor_select_region( editor,
					     found_at, length );
	
		g_object_set_data( G_OBJECT( widget ), "length",
				   GINT_TO_POINTER( length ) );
	} else {
		/* inform user not found */

	}

	g_free( text );
}

/**
 * do_replace:
 *
 * @widget: the replace button
 *
 * replaces the last found word
 *
 * returns: none
 */
void do_replace( GtkWidget *widget )
{
	GladeXML *xml;
	GtkWidget *entry;
	const gchar *replace_string;

	gint found_at;
	gint length;

	ScreemWindow *window;
	ScreemSite *site;
	ScreemPage *page;
	ScreemEditor *editor;

	xml = glade_get_widget_tree( widget );
	
	widget = glade_xml_get_widget( xml, "find_dialog" );
	window = SCREEM_WINDOW( g_object_get_data( G_OBJECT( widget ),
						   "window" ) );
	editor = SCREEM_EDITOR( window->details->editor );
	site = screem_window_get_current( window );

	page = screem_window_get_document( window );

	widget = glade_xml_get_widget( xml, "replace_entry" );
	entry = gnome_entry_gtk_entry( GNOME_ENTRY( widget ) );
	replace_string = gtk_entry_get_text( GTK_ENTRY( entry ) );

	found_at = GPOINTER_TO_INT( g_object_get_data( G_OBJECT( widget ), 
						       "found_at" ) );
	length = GPOINTER_TO_INT( g_object_get_data( G_OBJECT( widget ),
						     "length" ) );

	if( found_at == -1 ) {
		/* there was no string found */
		return;
	}

	screem_editor_delete_forward( editor, found_at, length );
	screem_editor_insert( editor, found_at, replace_string );
	found_at += strlen( replace_string );

	g_object_set_data( G_OBJECT( widget ), "found_at",
			   GINT_TO_POINTER( found_at ) );

	while( gtk_events_pending() ) {
		gtk_main_iteration();
	}

	screem_editor_set_pos( editor, found_at );

	/* find the next one automatically */
	do_find( widget );
}

/**
 * do_replace_all:
 *
 * @widget: the replace all button
 *
 * replaces all occurances of the find pattern
 *
 * returns: none
 */
void do_replace_all( GtkWidget *widget )
{
	GladeXML *xml;
	gint found_at;
	GSList *list;
	GtkWidget *widget2;

	ScreemWindow *window;
	ScreemPage *page;

	xml = glade_get_widget_tree( widget );

	
	widget = glade_xml_get_widget( xml, "find_dialog" );
	window = SCREEM_WINDOW( g_object_get_data( G_OBJECT( widget ),
						   "window" ) );

	widget = glade_xml_get_widget( xml, "match_list" );
	list = (GSList*)g_object_get_data( G_OBJECT( widget ), "matches" );

	widget = glade_xml_get_widget( xml, "replace_entry" );
	widget2 = glade_xml_get_widget( xml, "match_window" );

	found_at = 0;

	while( list || ( found_at != -1 ) ) {
		found_at = (gint) g_object_get_data( G_OBJECT( widget ), 
						     "found_at" );
		
		if( list ) {
			/* insert the next page */
			page = SCREEM_PAGE( list->data );
			screem_window_set_document( window, page );
			found_at = -1;
		}			

		if( found_at == -1 ) {
			do_find( widget );
		}

		found_at = (gint) g_object_get_data( G_OBJECT( widget ), 
						     "found_at" );

		/* replace all occurances in this document */
		while( found_at != -1 ) {
			do_replace( widget );
			found_at = (gint) g_object_get_data( G_OBJECT( widget ),
							     "found_at" );
		}

		while( gtk_events_pending() )
			gtk_main_iteration();

		if( list ) {
			list = list->next;
		}
       	}
}

void find_change_file( GtkTreeView *view, GtkTreePath *path,
		       GtkTreeViewColumn *col, gpointer data )
{
	ScreemWindow *window;
	ScreemSite *site;
	ScreemPage *page;
	GtkTreeModel *model;
	GtkTreeIter it;
	GValue value = { 0 };
	const gchar *pathname;

	window = SCREEM_WINDOW( data );
	
	model = gtk_tree_view_get_model( view );
	gtk_tree_model_get_iter( model, &it, path );
	gtk_tree_model_get_value( model, &it, 0,  &value );
	pathname = g_value_get_string( &value );

	site = screem_window_get_current( window );

	page = screem_site_locate_page( site, pathname );

	g_value_unset( &value );

	if( page ) {
		GladeXML *xml;
		GtkWidget *widget;

		xml = glade_get_widget_tree( GTK_WIDGET( view ) );
		widget = glade_xml_get_widget( xml, "replace_entry" );

		screem_window_set_document( window, page );
		
		/* open the page in the editor and perform a single page search */
		while( gtk_events_pending() )
			gtk_main_iteration();
		
		g_object_set_data( G_OBJECT( widget ), "found_at",
				   GINT_TO_POINTER( -1 ) );
		
		do_find( widget );
	}
}

/**
 * search:
 *
 * @text:      the text to search
 * @find:      the string to find
 * @use_regex: search uses regular expressions
 * @start:     offset in string to start searching at
 * @length:    a pointer to an integer to store the length of the match in,
 *             can be NULL if the value isn't interested in
 *
 * finds a string inside a piece of text
 *
 * returns: the position in text that find was found
 */
static gint search( const gchar *text, const gchar *find, gboolean use_regex,
		    gint start, gint *length )
{
	gint retval = start;

	g_assert( length != NULL );

	if( use_regex ) {
		regex_t expressions;
		size_t nmatch = 1;
		regmatch_t pmatch[ 1 ];
		
		if( regcomp( &expressions, find, REG_EXTENDED | REG_NEWLINE ) ||
		    regexec( &expressions, text + start, nmatch, pmatch, 0 ) ) {
			retval = -1;
		} else {
			 retval += pmatch[ 0 ].rm_so;
		}

		*length = pmatch[0].rm_eo - pmatch[0].rm_so;
	} else {
		/* not using regular expressions */
		gchar *pos;

		pos = strstr( text + start, find );
		if( ! pos ) {
			retval = -1;
		} else {
			retval = pos - text;
		}
		
		*length = strlen( find );
	}

	return retval;
}
