/*  Screem:  screem-page-model.c
 *
 *  The ScreemPageModel object
 *  
 *  This object is responsible for keeping a tree view of a ScreemPage
 *  up to date with changes made to the GtkTextBuffer of the page
 *  or at least that is the intention.
 *
 *  It will also be responsible for keeping track of the DTD in use
 *  for the page rather than the horrible method of the ScreemWindow
 *  handling this in combination with ScreemPage
 *
 *  The problem here though is coming up with an algorithm for
 *  dynamically updating the tree on text changes
 * 
 *  Copyright (C) 2003 David A Knight
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  For contact information with the author of this source code please see
 *  the AUTHORS file.  If there is no AUTHORS file present then check the
 *  about box under the help menu for a contact address
 */

#include <config.h>

#include <string.h>

#include "screem-application.h"

#include "screem-page-model.h"

#include "screem-dtd.h"
#include "screem-markup.h"
#include "screem-page.h"

#include "editMenu.h"

static void screem_page_model_class_init( ScreemPageModelClass *klass );
static void screem_page_model_init( ScreemPageModel *page );
static void screem_page_model_finalize( GObject *object );
static void screem_page_model_set_prop( GObject *object, guint prop_id,
				  	const GValue *value, GParamSpec *spec );
static void screem_page_model_get_prop( GObject *object, guint prop_id,
				  	GValue *value, GParamSpec *spec );

static void screem_page_model_page_set( ScreemPageModel *model );
static void insert_text( GtkTextBuffer *buffer, GtkTextIter *it,
			 const gchar *text, gint length, ScreemPageModel *model );
static void delete_range( GtkTextBuffer *buffer, GtkTextIter *it,
			  GtkTextIter *eit, ScreemPageModel *model );

static void screem_page_doctype_check( ScreemPageModel *model, gint pos, 
				       gint length );
static void screem_page_charset_check( ScreemPageModel *model, gint pos, 
				       gint length );

static void screem_page_model_dtd_set( ScreemPageModel *model );

typedef enum {
	PROP_0,
	PROP_APPLICATION,
	PROP_PAGE,
	PROP_DTD,
	PROP_CHARSET,
	PROP_DOCTYPE
} ScreemPageModelProps;

struct ScreemPageModelPrivate {
	ScreemApplication *application;
	
	ScreemPage *page;

	ScreemDTD *dtd;

	gint doctype_pos;
	gchar *doctype;

	gint charset_pos;
	gint charset_len;
	gchar *charset;
};

ScreemPageModel *screem_page_model_new( ScreemPage *page, GObject *app )
{
	ScreemPageModel *model;
	GType type;

	type = screem_page_model_get_type();

	model = SCREEM_PAGE_MODEL( g_object_new( type, 
						 "page", page,
						 "application", app,
						 NULL ) );

	return model;
}

void screem_page_model_force_check( ScreemPageModel *model )
{
	if( screem_page_is_markup( model->private->page ) ) {
		screem_page_doctype_check( model, 0, 0x7fffffff );
		
		screem_page_charset_check( model, 0, 0x7fffffff );
	}
}

/* static stuff */
static void screem_page_model_page_set( ScreemPageModel *model )
{
	GtkTreeIter it;
	GtkTextBuffer *buffer;

	gtk_tree_store_clear( GTK_TREE_STORE( model ) );
	
	gtk_tree_store_append( GTK_TREE_STORE( model ), &it, NULL );
	gtk_tree_store_set( GTK_TREE_STORE( model ), &it,
			    SCREEM_PAGE_MODEL_NAME, "[Document]",
			    SCREEM_PAGE_MODEL_VALID, TRUE,
			    -1 );

	buffer = screem_page_get_text_buffer( model->private->page );

	g_signal_connect_after( G_OBJECT( buffer ), "insert_text",
			  G_CALLBACK( insert_text ), model );
	g_signal_connect_after( G_OBJECT( buffer ), "delete_range",
				G_CALLBACK( delete_range ), model );

}

static void insert_text( GtkTextBuffer *buffer, GtkTextIter *it,
			 const gchar *text, gint length, ScreemPageModel *model )
{
	guint pos;

	/* - length as we are connected after, so the iter will be advanced
	 * by length chars */
	pos = gtk_text_iter_get_offset( it ) - length + 1;

	/* hmm, can end up with - pos here, possibly due to the initial
	 * newline chars in the buffer? */
	if( pos < 0 ) {
		pos = 0;
	}
	
	if( screem_page_is_markup( model->private->page ) ) {
		screem_page_doctype_check( model, pos, length );
		screem_page_charset_check( model, pos, length );
	}
}

static void delete_range( GtkTextBuffer *buffer, GtkTextIter *it,
			  GtkTextIter *eit, ScreemPageModel *model )
{
	guint pos;
	guint length;
	
	pos = gtk_text_iter_get_offset( it );
	length = gtk_text_iter_get_offset( eit ) - pos;
	
	if( screem_page_is_markup( model->private->page ) ) {
		screem_page_doctype_check( model, pos, length );
		screem_page_charset_check( model, pos, length );
	}

}

static void screem_page_doctype_check( ScreemPageModel *model, gint pos, 
				       gint length )
{
	ScreemPageModelPrivate *private;
	gboolean set;

	private = model->private;

	set = FALSE;

	if( ! private->doctype ) {
		set = TRUE;
	} else { 
		gint len;
		
		len = strlen( private->doctype );
			
		if( ( pos <= ( private->doctype_pos + len ) ) &&
		    ( pos + length ) >= private->doctype_pos ) {
			set = TRUE;
		}
	}
	if( set ) {
		/* set the doctype */
		gchar *text;
		gchar *doctype;
		
		text = screem_page_get_data( private->page );
		
		doctype = find_text( text,
				     "<![Dd][Oo][Cc][Ty][Yy][Pp][Ee] ",
				     NULL, NULL );
		if( doctype ) {
			GString *tag;
			
			tag = g_string_new( NULL );
			
			private->doctype_pos = doctype - text;
			
			doctype --;
			do {
				doctype ++;
				g_string_append_c( tag, *doctype );
			} while( *doctype != '\0' && *doctype != '>' );
		
			g_object_set( G_OBJECT( model ), "doctype", tag->str, NULL );
			
			g_string_free( tag, TRUE );
		}
			
		g_free( text );
	}
}

static void screem_page_charset_check( ScreemPageModel *model, gint pos, 
				       gint length )
{
	ScreemPageModelPrivate *private;
	gboolean set = FALSE;

	private = model->private;

	if( ! private->charset ) {
		set = TRUE;
	} else { 
		gint len;

		len = private->charset_len;

		if( ( pos <= ( private->charset_pos + len ) ) &&
		    ( pos + length ) >= private->charset_pos ) {
			set = TRUE;
		}
	}
	if( set ) {
		/* set the charset */
		gchar *text;
		gchar *tag;
		gchar *name;
		gint pos;
		gint start;
		gint end;
		gint len;

		pos = 0;
		text = screem_page_get_data( private->page );
		len = strlen( text );
		
		g_free( private->charset );
		private->charset = NULL;
		private->charset_pos = 0;

		/* first we look for <?xml encoding="" ?>
		   then we check <meta http-equiv="content-type" />
		   then we check <meta name="charset" /> 
		   if we have an <?xml ?> without an encoding="" we
		   enforce UTF-8 no matter what <meta /> exist */

		while( ( tag = screem_markup_next_tag( text, pos,
						       &start, &end, 
						       &name ) ) ) {
			g_assert( name );
			pos = end;
			if( ! strcmp( "?xml", name ) ) {
				/* xml PI element */
				GSList *list;
				GSList *attr;

				list=screem_markup_build_attributes_list(tag,
									 NULL);
				for( attr = list; attr; attr = attr->next ) {
					if(! strcmp( "encoding", 
						     attr->next->data ) &&
					   attr->data ) {
						g_object_set( G_OBJECT( model ), "charset", attr->data, NULL );
						break;
					} else {
						attr = NULL;
						break;
					}
					attr = attr->next;
				}
				
				if( ! attr ) {
					private->charset_pos = start;
					private->charset_len = end - start;
					g_object_set( G_OBJECT( model ), "charset", "UTF-8", NULL );
				}

				g_slist_foreach( list, (GFunc)g_free, NULL );
				g_slist_free( list );
				pos = len;
			} else if( ! g_strcasecmp( "meta", name ) ) {
				/* meta element */
				GSList *list;
				GSList *attr;
				gboolean content;
				
				list=screem_markup_build_attributes_list(tag,
									 NULL);
				content = FALSE;

				for( attr = list; attr; attr = attr->next ) {
					if( ! strcmp( "http-equiv", 
						      attr->next->data ) && 
					    attr->data &&
					    ! strcmp( "content-type",
						      attr->data ) ) {
						content = TRUE;
						break;
					} else if(! strcmp( "name", 
							    attr->next->data) &&
						  attr->data && 
						  ! strcmp( "charset",
							    attr->data)){
						break;
					}
					attr = attr->next;
				}
				if( attr ) {
					for( attr = list; attr; 
					     attr = attr->next ) {
						if( ! strcmp( "content",
							      attr->next->data ) ) {
							break;
						}
						attr = attr->next;
					}
				}
				if( attr && attr->data && content ) {
					gchar *val;

					val = strstr( attr->data, "charset=" );
					if( val ) {
						GString *tmp;

						val += strlen( "charset=" );
						tmp = g_string_new( NULL );
						while( *val != ' ' &&
						       *val != '\0' &&
						       *val != '"' ) {
							g_string_append_c( tmp,
									   *val);
							val ++;
						}
						private->charset_pos = start;
						private->charset_len = end - 
							start;
						pos =  len;
						g_object_set( G_OBJECT( model ), "charset", tmp->str, NULL );
						g_string_free( tmp, TRUE );
					}
				} else if( attr ) {
					private->charset_pos = start;
					private->charset_len = end - start;
					pos = len;
					g_object_set( G_OBJECT( model ), "charset", attr->data, NULL );
				}

				g_slist_foreach( list, (GFunc)g_free, NULL );
				g_slist_free( list );
			} else if( g_strcasecmp( "html", name ) &&
				   g_strcasecmp( "head", name ) &&
				   g_strcasecmp( "title", name ) &&
				   g_strcasecmp( "link", name ) &&
				   g_strcasecmp( "script", name ) &&
				   g_strcasecmp( "style", name ) &&
				   g_strcasecmp( "!--", name ) &&
				   g_strcasecmp( "base", name ) &&
				   g_strcasecmp( "isindex", name ) &&
				   g_strcasecmp( "!DOCTYPE", name ) &&
				   *name != '/' && *name != '?' ) {
				/* limiting case, bit of a hack for html,
				   valid xml docs are supposed to have the
				   <?xml ?> element and have indeterminate 
				   element names so we can't stop if the doc is
				   invalid */
				pos = len;
			}
			g_free( tag );
			g_free( name );
		}
	
		g_free( text );
	}
}

static void screem_page_model_dtd_set( ScreemPageModel *model )
{
	ScreemPageModelPrivate *priv;
	ScreemDTD *dtd;
	
	g_return_if_fail( SCREEM_IS_PAGE_MODEL( model ) );
	
	priv = model->private;
	
	if( priv->doctype ) {
		dtd = screem_application_add_dtd( priv->application, 
						  priv->doctype, NULL );
	} else {
		dtd = NULL;
	}

	if( ! dtd ) {
		dtd = screem_application_get_default_dtd( priv->application );
	}

	if( dtd != priv->dtd ) {
		g_object_set( G_OBJECT( model ), "dtd", dtd, NULL );
	}
}


/* G Object stuff */
#define PARENT_TYPE GTK_TYPE_TREE_STORE

static gpointer parent_class;

static void screem_page_model_class_init( ScreemPageModelClass *klass )
{
	GObjectClass *object_class;

	object_class = G_OBJECT_CLASS( klass );
	parent_class = g_type_class_peek_parent( klass );

	object_class->finalize = screem_page_model_finalize;
	object_class->get_property = screem_page_model_get_prop;
	object_class->set_property = screem_page_model_set_prop;

	g_object_class_install_property(object_class,
					PROP_APPLICATION,
					g_param_spec_object("application",
							    "Application",
							    "The Application",
							    G_TYPE_OBJECT,
							    G_PARAM_READWRITE|
							    G_PARAM_CONSTRUCT)
					);
	g_object_class_install_property(object_class,
					PROP_PAGE,
					g_param_spec_object("page",
							    "Page",
							    "The Page",
							    G_TYPE_OBJECT,
							    G_PARAM_READWRITE)
					);
	g_object_class_install_property(object_class,
					PROP_DTD,
					g_param_spec_object("dtd",
							    "DTD",
							    "The page DTD",
							    G_TYPE_OBJECT,
							    G_PARAM_READWRITE)
					);
	g_object_class_install_property(object_class,
					PROP_CHARSET,
					g_param_spec_string("charset",
							    "Page character set",
							    "The character set of the page",
							    "",
							    G_PARAM_READWRITE)
					);
	g_object_class_install_property(object_class,
					PROP_DOCTYPE,
					g_param_spec_string("doctype",
							    "Page doctype",
							    "The doctype of the page",
							    "",
							    G_PARAM_READWRITE)
					
					);

}

static void screem_page_model_init( ScreemPageModel *model )
{
	GConfClient *client;
	GType model_types[] = {
		G_TYPE_STRING,
		G_TYPE_STRING,
		G_TYPE_UINT,
		G_TYPE_UINT,
		G_TYPE_BOOLEAN,
		G_TYPE_UINT
	};
	GtkTreeIter it;
	
	model->private = g_new0( ScreemPageModelPrivate, 1 );

	client = gconf_client_get_default();
	model->private->charset = 
		gconf_client_get_string( client,
					"/apps/screem/editor/default_charset",
					NULL );
	g_object_unref( client );

	gtk_tree_store_set_column_types( GTK_TREE_STORE( model ),
					 SCREEM_PAGE_MODEL_MAX,
					 model_types );

	gtk_tree_store_append( GTK_TREE_STORE( model ), &it, NULL );
	gtk_tree_store_set( GTK_TREE_STORE( model ), &it,
			    SCREEM_PAGE_MODEL_NAME, "[Document]",
			    SCREEM_PAGE_MODEL_VALID, TRUE,
			    -1 );
}

static void screem_page_model_finalize( GObject *object )
{
	ScreemPageModel *model;

	model = SCREEM_PAGE_MODEL( object );

	g_free( model->private->charset );
	g_free( model->private->doctype );

	g_free( model->private );

	G_OBJECT_CLASS( parent_class )->finalize( object );
}

static void screem_page_model_set_prop( GObject *object, guint prop_id,
					const GValue *value, GParamSpec *spec )
{
	ScreemPageModel *model;
	const gchar *str;
	
	model = SCREEM_PAGE_MODEL( object );

	switch( prop_id ) {
		case PROP_APPLICATION:
			model->private->application = SCREEM_APPLICATION( g_value_get_object( value ) );
			model->private->dtd = screem_application_get_default_dtd( model->private->application );
			break;
		case PROP_PAGE:
			model->private->page = SCREEM_PAGE( g_value_get_object( value ) );
			screem_page_model_page_set( model );
			break;
		case PROP_DTD:
			model->private->dtd = SCREEM_DTD( g_value_get_object( value ) );
			g_object_set( G_OBJECT( model->private->page ), "dtd",
					model->private->dtd, NULL );
			break;
		case PROP_CHARSET:
			g_free( model->private->charset );
			model->private->charset = NULL;
			str = g_value_get_string( value );
			if( str ) {
				model->private->charset = g_strdup( str );
			}
			break;
		case PROP_DOCTYPE:
			g_free( model->private->doctype );
			model->private->doctype = NULL;
			str = g_value_get_string( value );
			if( str ) {
				model->private->doctype = g_strdup( str );
			}
			screem_page_model_dtd_set( model );
			break;
		default:
			g_warning( "Unknown property\n" );
	}
}

static void screem_page_model_get_prop( GObject *object, guint prop_id,
				  	GValue *value, GParamSpec *spec )
{
	ScreemPageModel *model;

	model = SCREEM_PAGE_MODEL( object );

	switch( prop_id ) {
		case PROP_APPLICATION:
			g_value_set_object( value, model->private->application );
			break;
		case PROP_PAGE:
			g_value_set_object( value, model->private->page );
			break;
		case PROP_DTD:
			g_value_set_object( value, model->private->dtd );
			break;
		case PROP_CHARSET:
			if( model->private->charset ) {
				g_value_set_string( value, model->private->charset );
			}
			break;
		case PROP_DOCTYPE:
			if( model->private->doctype ) {
				g_value_set_string( value, model->private->doctype );
			}
			break;
		default:
			break;
	}
}

GType screem_page_model_get_type()
{
	static GType type = 0;
	
	if( ! type ) {
		static const GTypeInfo info = {
			sizeof( ScreemPageModelClass ),
			NULL, /* base init */
			NULL, /* base finalise */
			(GClassInitFunc)screem_page_model_class_init,
			NULL, /* class finalise */
			NULL, /* class data */
			sizeof( ScreemPageModel ),
			0, /* n_preallocs */
			(GInstanceInitFunc)screem_page_model_init
		};

		type = g_type_register_static( PARENT_TYPE,
					       "ScreemPageModel",
					       &info, 0 );
	}

	return type;
}
